<?php

namespace Drupal\private_message\Entity;

use Drupal\Core\Cache\Cache;
use Drupal\Core\Entity\ContentEntityBase;
use Drupal\Core\Entity\EntityStorageInterface;
use Drupal\Core\Entity\EntityTypeInterface;
use Drupal\Core\Field\BaseFieldDefinition;
use Drupal\Core\Field\FieldStorageDefinitionInterface;
use Drupal\private_message\Entity\PrivateMessageInterface;
use Drupal\private_message\Entity\PrivateMessageThreadAccessTime;
use Drupal\private_message\Entity\PrivateMessageThreadDeleteTime;
use Drupal\user\UserInterface;

 /**
  * @ContentEntityType(
  *   id = "private_message_thread",
  *   label = @Translation("Private Message Thread"),
  *   handlers = {
  *     "view_builder" = "Drupal\private_message\Entity\Builder\PrivateMessageThreadViewBuilder",
  *     "views_data" = "Drupal\views\EntityViewsData",
  *     "access" = "Drupal\private_message\Entity\Access\PrivateMessageThreadAccessControlHandler",
  *   },
  *   base_table = "private_message_threads",
  *   admin_permission = "administer private messages",
  *   fieldable = TRUE,
  *   entity_keys = {
  *     "id" = "id",
  *     "uuid" = "uuid"
  *   },
  *   links = {
  *     "canonical" = "/private_messages/{private_message_thread}",
  *   },
  *   field_ui_base_route = "private_message.private_message_thread_settings",
  * )
  */
class PrivateMessageThread extends ContentEntityBase implements PrivateMessageThreadInterface {

  /**
   * {@inheritdoc}
   */
  public function addMember(UserInterface $user) {
    $this->get('members')->appendItem($user->id());
    $this->addLastAccessTime($user);
    $this->addLastDeleteTime($user);

    return $this;
  }

  /**
   * {@inheritdoc}
   */
  public function addMemberById($id) {
    $this->get('members')->appendItem($id);

    return $this;
  }

  /**
   * {@inheritdoc}
   */
  public function getMembers() {
    return $this->get('members')->referencedEntities();
  }

  /**
   * @inheritdoc}
   */
  public function isMember($id) {
    $members = $this->getMembers();
    foreach ($members as $member) {
      if ($member->id() == $id) {
        return TRUE;
      }
    }

    return FALSE;
  }

   /**
   * {@inheritdoc}
   */
  public function addMessage(PrivateMessageInterface $privateMessage) {
    $this->get('private_messages')->appendItem($privateMessage->id());

    return $this;
  }

  /**
   * {@inheritdoc}
   */
  public function addMessageById($id) {
    $this->get('private_messages')->appendItem($id);

    return $this;
  }

  /**
   * {@inheritdoc}
   */
  public function getMessages() {
    return $this->get('private_messages')->referencedEntities();
  }

  /**
   * {@inheritdoc}
   */
  public function filterNewestMessages() {
    $messages = $this->getMessages();
    if (count($messages) > \Drupal::config('private_message_thread.settings')->get('message_count')) {
      $list = $this->get('private_messages');
      $filtered_messages = array_slice($messages, -1 * \Drupal::config('private_message_thread.settings')->get('message_count'));
      $first_message = array_shift($filtered_messages);
      $first_key = $first_message->id();
      foreach ($list->referencedEntities() as $index => $list_item) {
        if ($list_item->id() < $first_key) {
          $list->removeItem(0);
        }
        else {
          break;
        }
      }
    }
  }

  /**
   * {@inheritdoc}
   */
  public function getNewestMessageCreationTimestamp() {
    $messages = $this->getMessages();
    $last_timestamp = 0;
    foreach ($messages as $message) {
      $creation_date = $message->get('created')->value;
      $last_timestamp = $creation_date > $last_timestamp ? $creation_date : $last_timestamp;
    }

    return $last_timestamp;
  }

  /**
   * {@inheritdoc}
   */
  public function addLastAccessTime(UserInterface $user) {
    $last_access_time = PrivateMessageThreadAccessTime::create([
      'owner' => $user->id(),
      'access_time' => REQUEST_TIME,
    ]);

    $this->get('last_access_time')->appendItem($last_access_time);
  }

  /**
   * {@inheritdoc}
   */
  public function getLastAccessTime(UserInterface $user) {
    $private_message_last_access = FALSE;
    foreach ($this->get('last_access_time') as $last_access_time) {
      if ($last_access_time->entity->getOwnerId() == $user->id()) {
        $private_message_last_access = $last_access_time;

        break;
      }
    }

    return $private_message_last_access;
  }

  /**
   * {@inheritdoc}
   */
  public function getLastAccessTimestamp(UserInterface $user) {
    $last_access = $this->getLastAccessTime($user);

    return $last_access ? $last_access->entity->get('access_time')->value : FALSE;
  }

  /**
   * {@inheritdoc}
   */
  public function updateLastAccessTime(UserInterface $user) {
    $private_message_access_time = $this->getLastAccessTime($user);

    if ($private_message_access_time) {
      $private_message_access_time->entity->setAccessTime(REQUEST_TIME)->save();
    }
    else {
      $this->addLastAccessTime($user);
    }

    return $this;
  }

  /**
   * {@inheritdoc}
   */
  public function addLastDeleteTime(UserInterface $user) {
    $last_delete_time = PrivateMessageThreadDeleteTime::create([
      'owner' => $user->id(),
      'delete_time' => 0,
    ]);

    $this->get('last_delete_time')->appendItem($last_delete_time);
  }

  /**
   * {@inheritdoc}
   */
  public function getLastDeleteTime(UserInterface $user) {
    $private_message_delete_time = FALSE;
    foreach ($this->get('last_delete_time') as $last_delete_time) {
      if ($last_delete_time->entity->getOwnerId() == $user->id()) {
        $private_message_delete_time = $last_delete_time->entity->get('delete_time')->value;

        break;
      }
    }

    return $private_message_delete_time;
  }

  /**
   * {@inheritdoc}
   */
  public function getLastDeleteTimes() {
    return $this->get('last_delete_time')->referencedEntities();
  }

  /**
   * {@inheritdoc}
   */
  public function updateLastDeleteTime(UserInterface $user) {
    $private_message_delete_time = $this->getLastDeleteTime($user);

    if ($private_message_delete_time) {
      $private_message_delete_time->setDeleteTime(REQUEST_TIME);
      $private_message_delete_time->save();
    }
    else {
      $this->addLastDeleteTime($user);
    }
  }

  /**
   * {@inheritdoc}
   */
  public function save() {
    parent::save();

    $tags = [];

    foreach ($this->getMembers() as $member) {
      $tags[] = 'private_message_inbox_block:uid:' . $member->id();
      $tags[] = 'private_message_notification_block:uid:' . $member->id();
    }

    Cache::invalidateTags($tags);
  }

  /**
   * {@inheritdoc}
   */
  public function delete(UserInterface $user = NULL) {
    if ($user) {
      $this->updateLastDeleteTime($user);
      $last_creation_timestamp = $this->getNewestMessageCreationTimestamp();
      $found = FALSE;
      $last_delete_times = $this->getLastDeleteTimes();
      foreach ($last_delete_times as $last_delete_time) {
        if ($last_delete_time->getDeleteTime() >= $last_creation_timestamp) {
          $found = TRUE;

          break;
        }
      }

      if (!$found) {
        parent::delete();
      }
    }
    else {
      parent::delete();
    }
  }

  /**
   * {@inheritdoc}
   */
  public function filterUserDeletedMessages(UserInterface $user) {
    $last_delete_timestamp = $this->getLastDeleteTime($user);
    $list = $this->get('private_messages');
    $messages = $this->getMessages();
    foreach ($messages as $index => $message) {
      if ($message->getCreatedTime() > $last_delete_timestamp) {
        break;
      }
    }
    return array_slice($messages, $index);
  }

  /**
   * {@inheritdoc}
   */
  public function getUpdatedTime() {
    return $this->get('updated')->value;
  }

  /**
   * {@inheritdoc}
   */
  public static function baseFieldDefinitions(EntityTypeInterface $entityType) {
    $fields = parent::baseFieldDefinitions($entityType);

    $fields['id']->setLabel(t('Custom private message thread ID'))
      ->setDescription(t('The private message thread ID.'));

    $fields['uuid']->setDescription(t('The custom private message thread UUID.'));

     $fields['updated'] = BaseFieldDefinition::create('changed')
      ->setLabel(t('Updated'))
      ->setDescription(t('The most recent time at which the thread was updated'));

   // Member(s) of the private message thread
    // Entity reference field, holds the reference to user objects.
    $fields['members'] = BaseFieldDefinition::create('entity_reference')
      ->setLabel(t('Member(s)'))
      ->setDescription(t('The member(s) of the private message thread'))
      ->setSetting('target_type', 'user')
      ->setSetting('handler', 'default')
      ->setCardinality(FieldStorageDefinitionInterface::CARDINALITY_UNLIMITED)
      ->addConstraint('private_message_thread_member')
      ->setDisplayOptions('view', array(
        'label' => 'above',
        'type' => 'entity_reference_label',
        'weight' => -3,
      ))
      ->setDisplayOptions('form', array(
        'type' => 'entity_reference_autocomplete',
        'settings' => array(
          'match_operator' => 'CONTAINS',
          'size' => 60,
          'autocomplete_type' => 'tags',
          'placeholder' => '',
        ),
        'weight' => -3,
      ))
      ->setDisplayConfigurable('form', TRUE)
      ->setDisplayConfigurable('view', TRUE);

    // Private messages in this thread
    // Entity reference field, holds the reference to user objects.
    $fields['private_messages'] = BaseFieldDefinition::create('entity_reference')
      ->setLabel(t('Messages'))
      ->setDescription(t('The private messages that belong to this thread'))
      ->setSetting('target_type', 'private_message')
      ->setSetting('handler', 'default')
      ->setCardinality(FieldStorageDefinitionInterface::CARDINALITY_UNLIMITED)
      ->setDisplayOptions('view', array(
        'label' => 'hidden',
        'type' => 'entity_reference_entity_view',
      ))
      ->setDisplayConfigurable('view', TRUE);

    // Last access timestamps for each member, representing the last
    // time at which they accessed the private message thread
    $fields['last_access_time'] = BaseFieldDefinition::create('entity_reference')
      ->setLabel(t('Last Access TImes'))
      ->setDescription(t('Timestamps at which members of this private message thread last accessed the thread'))
      ->setSetting('target_type', 'pm_thread_access_time')
      ->setSetting('handler', 'default')
      ->setCardinality(FieldStorageDefinitionInterface::CARDINALITY_UNLIMITED);

    // Last delete - timestamps for each member, representing the last
    // time at which they deleted the private message thread. Note that
    // the thread is only deleted from the database if/when all members have
    // deleted the thread. Until that point, this value is used to determine the first
    // message to be shown to a user if they have deleted the thread, then re-entered
    // a conversation with the other members of the thread.
    $fields['last_delete_time'] = BaseFieldDefinition::create('entity_reference')
      ->setLabel(t('Last Delete Times'))
      ->setDescription(t('Timestamps at which members of this private message thread last deleted the thread'))
      ->setSetting('target_type', 'pm_thread_delete_time')
      ->setSetting('handler', 'default')
      ->setCardinality(FieldStorageDefinitionInterface::CARDINALITY_UNLIMITED);

    return $fields;
  }

  /**
   * {@inheritdoc}
   */
  public function getCacheTags() {
    $tags = parent::getCacheTags();
    $tags[] = 'private_message_thread:' . $this->id() . ':view:uid:' . \Drupal::currentUser()->id();

    return $tags;
  }
}
